import platform
import subprocess
from ..constants import *
from ..layers import *
from ..utils import *
from ..materials import *
import json
import gdsfactory as gf
from copy import deepcopy
# from time import time
import datetime
from math import cos, pi, sin
import os
import numpy as np

from sortedcontainers import SortedDict, SortedSet
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def setup(path, c, study, nres, center_wavelength,
          bbox_layer=LAYER.WAFER,
          zmargin2=None, zlims=None, core_layer=LAYER.WG,
          port_source_offset="auto", port_margin="auto",
          runs=[],  sources=[],
          layer_stack=LAYER_STACK, materials=dict(),
          exclude_layers=[
              DESIGN_LAYER, GUESS], Courant=None,
          gpu=None, dtype=np.float32,
          plot=False, framerate=0,
          magic="", wd=os.path.join(os.getcwd(), "runs"), name=None,
          Ttrans=None,
          approx_2D_mode=False):
    RATIO = 4

    materials = {**MATERIALS, **materials}
    prob = dict()
    if approx_2D_mode:
        N = 2
        prob["approx_2D_mode"] = approx_2D_mode
    else:
        N = 3
        prob["approx_2D_mode"] = None
    dy = dx = center_wavelength/nres
    dl = dx/RATIO
    dz = 1 * dx

    prob["class"] = "pic"
    prob["Ttrans"] = Ttrans
    prob["path"] = path
    prob["name"] = name
    prob["center_wavelength"] = center_wavelength
    prob["dx"] = dx
    prob["dy"] = dy
    prob["dz"] = dz
    prob["dl"] = dl
    # ratio =
    # prob["ratio"] = ratio
    prob["dtype"] = str(dtype)
    prob["timestamp"] = datetime.datetime.now().isoformat(
        timespec="seconds").replace(":", "-")
    prob["magic"] = magic
    prob["framerate"] = framerate

    gpu_backend = gpu
    # if gpu_backend:s
    prob["gpu_backend"] = gpu_backend

    ports = {
        p.name: {
            "center": p.center,
            "normal": [cos(p.orientation/180*pi), sin(p.orientation/180*pi)],
        }
        for p in c.get_ports_list(prefix="o")
    }
    prob["ports"] = ports

    mode_solutions = []

    d = get_layers(layer_stack, core_layer)[0]
    hcore = d.thickness
    zcore = d.zmin

    zmargin1 = 3*hcore
    hmode = hcore+2*zmargin1
    hmode = trim(hmode, 2*dz)
    zmargin1 = (hmode-hcore)/2

    zmode = zcore-zmargin1
    zcenter = zcore+hcore/2

    zmargin2 = 2*hcore
    zmargin2 = trim(zmargin2,  2*dz)
    h = hcore+2*(zmargin1+zmargin2)
    zmin = zcore-zmargin2-zmargin1
    zmax = zmin+h

    source_ports = []
    nonsource_ports = []
    for p in c.ports:
        is_source = False
        for run in runs:
            for port in run["sources"]:
                if port == p.name:
                    is_source = True
        if is_source:
            source_ports.append(p.name)
        else:
            nonsource_ports.append(p.name)
#
    port_width = max([p.width/1e3 for p in c.ports])
    ps = portsides(c)
    xmargin = ymargin = 2*port_width

    source_port_margin = port_width if N == 2 else 6*port_width

    port_margin = 2*dx
    margins = []
    for p in ps:
        if set(p).intersection(source_ports):
            margins.append(source_port_margin+port_margin)
        elif set(p).intersection(nonsource_ports):
            margins.append(port_margin)
        else:
            assert not p
            margins.append(xmargin)
    l0, w0 = c.bbox_np()[1]-c.bbox_np()[0]

    _l = l0+margins[0]+margins[2]
    l = trim(_l, 2*dx)
    margins[0] += (l-_l)

    _w = w0+margins[1]+margins[3]
    w = trim(_w, 2*dy)
    margins[1] += (w-_w)

    #
    modemargin = 1*port_width
    wmode = port_width+2*modemargin
    wmode = trim(wmode, 2*dx)
    modemargin = (wmode-port_width)/2

    prob["hcore"] = hcore
    prob["zcenter"] = zcenter
    prob["zmin"] = zmin
    prob["zcore"] = zcore
    prob["xmargin"] = xmargin
    prob["ymargin"] = ymargin
    prob["zmargin2"] = zmargin2
    prob["L"] = [l, w, h]

    _c = gf.Component()
    kwargs = dict()
    for (orientation, side, length, p) in zip([0, 90, 180, 270], ["right", "top", "left", "bottom"], margins, ps):
        if p:
            c = gf.components.extend_ports(
                c, None, length, orientation=orientation)
        else:
            kwargs[side] = length
    _c << gf.components.bbox(component=c, layer=bbox_layer, **kwargs)
    c0 = _c << c
    c = _c
    # c.plot()
    # c.show()

    lb = c.bbox_np()[0].tolist()
    center = [c.bbox_np()[0][0], c.bbox_np()[0][1]+w/2, zcenter]

    xs = arange(lb[0], lb[0]+l, dx)
    ys = arange(lb[1], lb[1]+w, dy)
    # zs = sorted(list(set(
    #     arange(zmin, zcore-zmargin1, 4*dz) +
    #     arange(zcore-zmargin1, zcore-2*dz, 2*dz) +
    #     arange(zcore-2*dz, zcore+hcore+2*dz, dz) +
    #     arange(zcore+hcore+2*dz, zcore+hcore+zmargin1, 2*dz) +
    #     arange(zcore+hcore+zmargin1, zmax, 4*dz)
    # )))
    z0 = zmin
    z1 = zmode
    z2 = zmode+hmode
    z3 = zmax
    zs = sorted(list(set(
        arange(z0, z1, 2*dz) +
        arange(z1, z2, dz) +
        arange(z2, z3, 2*dz)
    )))
    prob["xs"] = xs
    prob["ys"] = ys
    prob["zs"] = zs

    for run in runs:
        for port in run["sources"]:
            s = c0.ports[port]
            a = pi/180*s.orientation
            center = np.array(s.center)/1e3 - \
                np.array([cos(a), sin(a)])*port_margin
            run["sources"][port]["center"] = center.tolist()

    layers = set(c.layers)-set(exclude_layers)

    normal = [1, 0, 0]

    temp = os.path.join(path, "geometry")
    os.makedirs(temp, exist_ok=True)
    layer_stack_info = material_voxelate(
        c, dl, zmin, zmax, layers, layer_stack, temp)
    dir = os.path.dirname(os.path.realpath(__file__))
    print(dir)
    fn = os.path.join(dir, "solvemodes.py")
    assert os.path.exists(fn)
    assert os.path.exists(temp)
    if platform.system() == "Windows":
        os.system(f"copy /Y {fn} {temp}")
    else:
        subprocess.run(["cp", fn, temp])
    prob["layer_stack"] = layer_stack_info
    # materials = set([v["material"] for v in layer_stack_info.values()])
    # d = {
    #     k: materials[k] for k in materials
    # }
    prob["study"] = study
    # mateps = {k: np.real(v) for k, v in d.items()}
    # materials = {
    #     k: {} for k in materials
    # }
    # for k, v in mateps.items():
    #     materials[k]["epsilon"] = v
    prob["materials"] = materials

    prob["N"] = N

    prob["hmode"] = hmode
    prob["wmode"] = wmode
    prob["zmode"] = zmode
    wmode = port_width+2*modemargin
    # print(margin)
    neffmin = 1000000
    wavelengths = []
    # _c = add_bbox(c, layer=bbox_layer, nonport_margin=margin)
    for run in runs:
        for s in list(run["sources"].values())+list(run["monitors"].values()):
            s["mode_width"] = wmode
            for center_wavelength in s["wavelength_mode_numbers"]:
                wavelengths.append(center_wavelength)
    wavelengths = sorted(set(wavelengths))

    bbox = c.bbox_np()

    prob["mode_solutions"] = mode_solutions
    prob["runs"] = runs
    prob["components"] = {
        "device": {
            "bbox": c.bbox_np().tolist(),
        }}

    prob["Courant"] = Courant
    if not os.path.exists(path):
        os.makedirs(path)
    # prob["wavelengths"] = wavelengths
    c.write_gds(os.path.join(path, "component.gds"))
    prob["mode_solutions"] = mode_solutions
    # prob["path_length"] = 2*(l+r+lwg)
    return prob


def port_name(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        return s
    return f"o{s}"


def port_number(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        s = s[1:]
    return int(s)


def mode_number(port):
    l = str(port).split("@")
    return 0 if len(l) == 1 else int(l[1])


def unpack_sparam_key(k):
    o, i = k.split(",")
    po, pi = port_name(o), port_name(i)
    mo, mi = mode_number(o), mode_number(i)
    return po, mo, pi, mi


def long_sparam_key(k):
    po, mo, pi, mi = unpack_sparam_key(k)
    return f"{po}@{mo},{pi}@{mi}"



import time
import pymeshfix
import json
from statistics import median
from .constants import *
import gdsfactory as gf
from gdsfactory.cross_section import Section
from .constants import *
from .materials import *
try:
    from IPython.display import display
except ImportError:
    pass
import pyvista as pv

# from .picToGDS import main

from math import ceil, cos, pi, sin, tan
import matplotlib.pyplot as plt
import numpy as np
from gdsfactory.generic_tech import LAYER_STACK, LAYER
import copy
import shutil
# from gdsfactory import LAYER_VIEWS

tol = .001


def arange(a, b, d):
    ret = np.linspace(a, b, round((b-a)/(d))+1).tolist()
    return ret


def trim(x, dx):
    return round(x/dx)*dx


def extend(endpoints, wm):
    v = endpoints[1]-endpoints[0]
    v = v/np.linalg.norm(v)
    return [
        (endpoints[0]-wm*v).tolist(), (endpoints[1]+wm*v).tolist()]


def portsides(c):
    ports = c.ports
    bbox = c.bbox_np()
    res = [[], [], [], []]
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    for p in ports:
        x, y = np.array(p.center)/1e3

        if abs(x - xmin0) < tol:
            res[2].append(p.name)
        if abs(x - xmax0) < tol:
            res[0].append(p.name)
        if abs(y - ymin0) < tol:
            res[3].append(p.name)
        if abs(y - ymax0) < tol:
            res[1].append(p.name)
    return res


def add_bbox(c, layer, nonport_margin=0):  # , dx=None):
    bbox = c.bbox_np()
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    l = xmax0-xmin0
    w = ymax0-ymin0

    # l = dx*np.ceil((xmax0-xmin0)/dx)
    # w = dx*np.ceil((ymax0-ymin0)/dx)

    # if dx is not None:
    #     if nonport_margin is None:
    #         nonport_margin = dx
    # if nonport_margin is None:
    #     nonport_margin = 0
    margin = nonport_margin
    xmin, ymin, xmax, ymax = xmin0-margin, ymin0 - \
        margin, xmin0+l+margin, ymin0+w+margin

    for p in c.ports:
        # p = c.ports[k]
        x, y = np.array(p.center)/1e3
        if abs(x - xmin0) < tol:
            xmin = x
        if abs(x - xmax0) < tol:
            xmax = x
        if abs(y - ymin0) < tol:
            ymin = y
        if abs(y - ymax0) < tol:
            ymax = y
    p = [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
    _c = gf.Component()
    _c << c

    if type(layer[0]) is int:
        layer = [layer]
    for layer in layer:
        layer = tuple(layer)
        _c.add_polygon(p, layer=layer)
    for port in c.ports:
        _c.add_port(name=port.name, port=port)
    return _c

# def pic2gds(fileName, sizeOfTheCell, layerNum=1, isDither=False, scale=1):
    main(fileName, sizeOfTheCell, layerNum, isDither, scale)
    return "image.bmp", "image.gds"


def finish(c, name):
    c.add_label(name, position=c.bbox_np()[1])


def normal_from_orientation(orientation):
    return [cos(orientation/180*pi), sin(orientation/180*pi)]


def stl_to_array(mesh: pv.PolyData, dl: float, bbox):
    # x_min, x_max, y_min, y_max, z_min, z_max = mesh.bounds
    # x_min,  y_min, z_min = bbox[0]
    # x_max, y_max, z_max = bbox[1]
    # x = np.linspace(x_min+dl/2, x_max-dl/2, round((x_max-x_min)/dl))
    # y = np.linspace(y_min+dl/2, y_max-dl/2, round((y_max-y_min)/dl))
    # z = np.linspace(z_min+dl/2, z_max-dl/2, round((z_max-z_min)/dl))
    # x, y, z = np.meshgrid(x, y, z)

    lb = [mesh.bounds[i] for i in [0, 2, 4]]
    ub = [mesh.bounds[i] for i in [1, 3, 5]]
    lb = [max(a, b) for a, b in zip(lb, bbox[0])]
    ub = [min(a, b) for a, b in zip(ub, bbox[1])]

    lims = [
        [a+dl/2+dl*ceil((p-a-dl/2)/dl), b-dl/2-dl*ceil((b-dl/2-q)/dl)]
        for (a, b, p, q) in zip(bbox[0], bbox[1], lb, ub)]
    xyz = [np.linspace(a, b, 1+round((b-a)/dl)) for (a, b) in lims]
    x, y, z = np.meshgrid(*xyz)

    # Create unstructured grid from the structured grid
    grid = pv.StructuredGrid(x, y, z)
    ugrid = pv.UnstructuredGrid(grid)

    # Get part of the mesh within the mesh's bounding surface.
    selection = ugrid.select_enclosed_points(
        mesh.extract_surface(),
        tolerance=0.00,
        # check_surface=False,
    )
    mask = selection['SelectedPoints'].view(bool)
    mask = mask.reshape(x.shape, order='F')
    mask = np.array(mask)
    mask = np.transpose(mask, (1, 0, 2))

    start = [(lims[i][0]-a-dl/2)/dl for (i, a) in enumerate(bbox[0])]
    stop = [a+l for (a, l) in zip(start, mask.shape)]
    start = [int(a) for a in start]
    stop = [int(a) for a in stop]

    r = np.zeros(tuple([int((b-a)/dl)
                 for a, b in zip(bbox[0], bbox[1])]), dtype=bool)
    r[start[0]:stop[0], start[1]:stop[1], start[2]:stop[2]] = mask
    return r


def material_voxelate(c, dl, zmin, zmax, layers, layer_stack, path):
    stacks = sum([[[v.mesh_order, v.material, tuple(layer), k]
                 for k, v in get_layers(layer_stack, layer, withkey=True)] for layer in layers], [])
    c.flatten()
    stacks = sorted(stacks, key=lambda x: -x[0])
    layer_stack_info = dict()
    # c.show()
    # raise NotImplementedError("This is a stub")
    lb, ub = c.bbox_np()
    # bbox = [[**lb, zmin], [**ub, zmax]]
    bbox = [[lb[0], lb[1], zmin], [ub[0], ub[1], zmax]]
    for i, stack in enumerate(stacks):
        m = stack[1]
        l1, l2 = layer = stack[2]
        k = stack[3]

        _layer_stack = copy.deepcopy(layer_stack)
        _layer_stack.layers.clear()

        d = copy.deepcopy(layer_stack.layers[k])
        if d.zmin <= zmax and d.bounds[1] >= zmin:
            _layer_stack.layers[k] = d
            d.zmin = max(zmin, d.zmin)
            d.thickness = min(zmax-d.zmin, d.thickness)
            # _d.bounds = (_d.zmin, _d.zmin+_d.thickness)
            # origin = (c.extract([layer]).bbox_np()-c.bbox_np())[0].tolist()
            STL = os.path.abspath(os.path.join(path, f"{k}.stl"))
            gf.export.to_stl(c, STL, layer_stack=_layer_stack)
            STL = os.path.join(path, f'{k}_{l1}_{l2}.stl')

            pymeshfix.clean_from_file(STL, STL)
            mesh = pv.read(STL)
            im = stl_to_array(mesh, dl, bbox)
            np.save(os.path.join(path, f'{k}.npy'), im)

            layer_stack_info[k] = {
                "layer": (l1, l2),
                "zmin": d.zmin,
                "thickness": d.thickness,
                "material": matname(m),
                "mesh_order": stack[0],
                # "origin": origin,
            }
    return layer_stack_info


def get_layers(layer_stack, layer, withkey=False):
    r = []
    for k, x in layer_stack.layers.items():
        l = x.layer
        if hasattr(l, "layer"):
            if tuple(l.layer) == tuple(layer):
                if withkey:
                    x = k, x
                # print(layer, k, x)
                r.append(x)
    if r:
        return r

    for k, x in layer_stack.layers.items():
        l = x.derived_layer
        if hasattr(l, "layer"):
            if tuple(l.layer) == tuple(layer):
                if withkey:
                    x = k, x
                # print(layer, k, x)
                r.append(x)
    return r


def wavelength_range(center, bandwidth, length=3):
    f1 = 1/(center+bandwidth/2)
    f2 = 1/(center-bandwidth/2)
    hw = (f2-f1)/2
    f1 = 1/center-hw
    f2 = 1/center+hw
    return sorted([1/x for x in np.linspace(f1, f2, length).tolist()])


def adjust_wavelengths(wavelengths, center_wavelength=None, wl_res=.01):
    if len(wavelengths) == 1:
        return wavelengths, wavelengths[0], 1
    wavelengths = sorted(set(wavelengths))
    wavelengths0 = copy.deepcopy(wavelengths)
    if center_wavelength is None:
        center_wavelength = median(wavelengths)
    freqs0 = [center_wavelength/w for w in reversed(wavelengths)]
    nresfreq = round(1/min(np.diff([0]+freqs0)))

    while True:
        freqs = [round(f*nresfreq)/nresfreq for f in freqs0]
        wavelengths = [center_wavelength/f for f in reversed(freqs)]
        if max([abs(w-w0)/center_wavelength for w, w0 in zip(wavelengths, wavelengths0)]) <= wl_res:
            break
        nresfreq += 1
    print(
        f"wavelengths has been adjusted within `wl_res` to facilitate simulation:\n{wavelengths}")
    return wavelengths, center_wavelength, nresfreq


def save_prob(prob, path):
    path = os.path.abspath(path)
    bson_data = json.dumps(prob)
    # prob["component"] = c0

    path = prob["path"]
    if not os.path.exists(path):
        os.makedirs(path)
        #   compiling julia code...
        #   """)
    prob_path = os.path.join(path, "problem.json")
    print(prob_path)
    with open(prob_path, "w") as f:
        # Write the BSON data to the file
        f.write(bson_data)
    print("using simulation folder", path)



# import dill
from pprint import pprint
import os
import subprocess
import json
import numpy as np
import requests
from .pic.inverse_design import *
from .pic.sparams import *
from .utils import *
from .layers import *
from .constants import *
from PIL import Image


from subprocess import Popen, PIPE
URL = "http://127.0.0.1:8975"


def start_fdtd_server(url=URL):
    try:
        r = requests.get(url)
    except:
        print("starting julia fdtd server on localhost...")
        cmd = ["julia", "-e", "using Luminescent; start_fdtd_server()"]
        proc = Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        with proc:
            for line in proc.stdout:
                print(str(line.decode().strip()), flush=True)
            err_message = proc.stderr.read().decode()
            print(err_message)


def solve(path, dev=False):
    path = os.path.abspath(path)
    # prob = load_prob(path)

    # prob["action"] = "solve"
    # r = requests.post(f"{url}/local", json=prob)
    1
    # print(f"""
    #       using simulation folder {path}
    #       starting julia process...
    #       """)
    # start_time = time.time()

    def run(cmd):

        proc = Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # proc.wait()
        with proc:
            for line in proc.stdout:

                print(str(line.decode().strip()), flush=True)
            err_message = proc.stderr.read().decode()
            print(err_message)
    # if dev:
    #     env = r'using Pkg;Pkg.activate(raw"c:\Users\pxshe\OneDrive\Desktop\beans\Luminescent.jl\luminescent");'
    # else:
    #     env = '0;'
    # cmd = ["lumi", path]
    # try:
    # run(["Luminescent", path])
    # except:
    run(["julia", "-e", f'println(Base.active_project())'])
    print("no fdtd binaries found - starting julia session to compile fdtd code - will take 5 mins - can take a break and come back :) ...")

    prob = json.loads(open(os.path.join(path, "problem.json"), "rb").read())
    a = ['julia', '-e', ]
    gpu_backend = prob["gpu_backend"]
    _class = prob["class"]
    if gpu_backend == "CUDA":
        array = "cu"
        pkgs = ",CUDA"
    else:
        array = "Array"
        pkgs = ""

    b = [f'using Luminescent{pkgs};{_class}run(raw"{path}",{array})']
    run(a+b)

    # with Popen(cmd,  stdout=PIPE, stderr=PIPE) as p:
    #     if p.stderr is not None:
    #         for line in p.stderr:
    #             print(line, flush=True)
    # exit_code = p.poll()
    # subprocess.run()
    # print(f"julia simulation took {time.time()-start_time} seconds")
    # print(f"images and results saved in {path}")
    # sol = load_res(path=path)
    # return sol


def load_sparams(sparams):
    if "re" in list(sparams.values())[0]:
        return {k: v["re"]+1j*v["im"] for k, v in sparams.items()}
    return {center_wavelength: {k: (v["re"]+1j*v["im"])
                                for k, v in d.items()} for center_wavelength, d in sparams.items()}


def load_res(path, show=True):
    path = os.path.abspath(path)
    print(f"loading solution from {path}")
    prob = json.loads(open(os.path.join(path, "problem.json"), "rb").read())
    p = os.path.join(path, "solution.json")
    # sol = json.loads(p, "rb").read())["sol"]
    sol = json.loads(open(p).read())
    sol["S"] = load_sparams(sol["S"])
    sol["component"] = gf.import_gds(os.path.join(path, "component.gds"))
    if prob["study"] == "sparams":
        pass
    elif prob["study"] == "inverse_design":
        l = [np.array(d) for d in sol["optimized_designs"]]
        sol["optimized_designs"] = l
        print(f"loading optimized design regions at resolution {sol['dl']}")
        # for i, a in enumerate(l):
        #     path = f"optimized_design_region_{i+1}.png"
        #     Image.fromarray(np.flip(np.uint8((1-a)) * 255, 0),
        #                     'L').save(os.path.join(path, name))
        # pic2gds(os.path.join(
        #     path, name), sol["dx"])
        # c = apply_design(sol["component"],  sol)
        # sol["optimized_component"] = copy.deepcopy(c)
        # sol["optimized_component"] = c
        # c.write_gds(os.path.join(path, "optimized_component.gds"))

    if show:
        i = 1
        while True:
            p = os.path.join(path, f"run_{i}.png")
            if os.path.exists(p):
                img = Image.open(p)
                img.show()
                try:
                    display(img)
                except:
                    pass
                i += 1
            else:
                break

        _sol = {k: sol[k] for k in ["T", "dB", "S", "phasors", "path", ]}
        _sol["optimized_designs"] = "[[...]]"
        pprint(_sol)

    return sol


def finetune(path, iters, **kwargs):
    fn = os.path.join(path, "problem.json")
    prob = json.loads(open(fn, "rb").read())
    prob["iters"] = iters
    prob["path"] = path
    prob["restart"] = False
    prob = {**prob, **kwargs}
    with open(fn, 'w') as f:
        json.dump(prob, f)
    return solve(path)



import matplotlib.pyplot as plt
from math import floor
from statistics import mode
from .setup import *
from ..constants import *
from ..layers import *
from ..utils import *
from ..materials import *
import gdsfactory as gf
from copy import deepcopy
import matplotlib.pyplot as plt
import numpy as np

from gdsfactory.cross_section import Section
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def make_sim_prob(
    path,
    wavelengths=[],
    center_wavelength=None,
    frequencies=[],
    center_frequency=None,
        sources=[],
        monitors=[],
        nres=None,
        dx=None,
        layer_stack={},
        materials={},
        study="simulation",
        dtype="float32",
        Ttrans=None,
        Tss=None,
        wl_res=.01,
        gpu=None,
        margins=[[0, 0, 0], [0, 0, 0]],
):
    RATIO = 2

    materials = {**MATERIALS, **materials}

    if not wavelengths:
        if not center_frequency:
            center_frequency = mode(frequencies)

        wavelengths = [center_wavelength * center_frequency /
                       f for f in sorted(frequencies, reverse=True)]
    wavelengths, center_wavelength, T = adjust_wavelengths(
        wavelengths, center_wavelength, wl_res)

    if nres:
        dx = center_wavelength/nres
    dy = dz = dx
    dl = dx/RATIO

    GEOMETRY = os.path.join(path, "geometry")
    bbox = [[None, None, None], [None, None, None]]
    for material in os.listdir(GEOMETRY):
        for fn in os.listdir(os.path.join(GEOMETRY, material)):
            if fn.lower().endswith(".stl"):
                STL = os.path.join(GEOMETRY, material, fn)
                print(STL)
                # pymeshfix.clean_from_file(STL, STL)
                mesh = pv.read(STL)
                for (i, v, w) in zip(range(3), bbox[0], [mesh.bounds[i] for i in [0, 2, 4]]):
                    if v is None or w < v:
                        bbox[0][i] = w
                for (i, v, w) in zip(range(3), bbox[1], [mesh.bounds[i] for i in [1, 3, 5]]):
                    if v is None or w > v:
                        bbox[1][i] = w
    bbox[0] = (np.array(bbox[0])-np.array(margins[0])).tolist()
    bbox[1] = (np.array(bbox[1])+np.array(margins[1])).tolist()

    bbox[1] = [a+dx*floor((b-a)/dx) for (a, b) in zip(bbox[0], bbox[1])]

    print(bbox)
    for material in os.listdir(GEOMETRY):
        for fn in os.listdir(os.path.join(GEOMETRY, material)):
            if fn.lower().endswith(".stl"):
                STL = os.path.join(GEOMETRY, material, fn)
                mesh = pv.read(STL)
                # mesh.plot()
                im = stl_to_array(mesh, dl,  bbox)
                print(im.shape)

                # fig = plt.figure()
                # ax = fig.add_subplot(111, projection='3d')
                # ax.voxels(im, edgecolor='k')
                # plt.show()

                name = fn[:-4]
                np.save(os.path.join(GEOMETRY, material, f'{name}.npy'), im)

    L = (bbox[1][0]-bbox[0][0], bbox[1][1]-bbox[0][1], bbox[1][2]-bbox[0][2])
    xs = np.linspace(bbox[0][0], bbox[1][0], 1 +
                     round((bbox[1][0]-bbox[0][0])/dx)).tolist()
    ys = np.linspace(bbox[0][1], bbox[1][1], 1 +
                     round((bbox[1][1]-bbox[0][1])/dy)).tolist()
    zs = np.linspace(bbox[0][2], bbox[1][2], 1 +
                     round((bbox[1][2]-bbox[0][2])/dz)).tolist()
#  dtype, center_wavelength, dl, xs, ys, zs, study, layer_stack, sources, monitors, materials, L, Ttrans, Tss, wavelengths
    if not Tss:
        Tss = T if len(wavelengths) > 1 else None
    prob = {
        "class": "gen",
        "sources": sources,
        "monitors": monitors,
        "nres": nres,
        "study": study,
        'dtype': dtype,
        'center_wavelength': center_wavelength,
        'dl': dl,
        'dx': dx,
        'xs': xs,
        'ys': ys,
        'zs': zs,
        'layer_stack': layer_stack,
        'materials': materials,
        'L': L,
        'gpu_backend': gpu,
        'Ttrans': Ttrans,
        'Tss': Tss,
    }
    prob["wavelengths"] = wavelengths
    PROB = os.path.join(path, "problem.json")
    with open(PROB, 'w') as f:
        json.dump(prob, f)
    return prob

    # l = [k for k in imow if port_number(k) == pi]
    # if not l:
    #     imow[f"o{pi}@{mi}"] = []
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[i] = imow[k]
    #         del imow[k]

    # l = [k for k in imow[i] if port_number(k) == po]
    # if not l:
    #     imow[f"o{pi}@{mi}"]
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[f"o{pi}@{mn}"] = imow[k]
    #         del imow[k]

    # if po not in imow[pi]:
    #     imow[pi]["o"][po] = mo
    # else:
    #     imow[pi]["o"][po] = max(imow[pi]["o"][po], mo)
